DEFAULT_BLOCKS_ARGS = {
    'efficientnetv2-s': [
        {
            'kernel_size': (3, 3),
            'num_repeat': 2,
            'features_in': 24,
            'features_out': 24,
            'expand_ratio': 1,
            'se_ratio': 0.0,
            'strides': 1,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 4,
            'features_in': 24,
            'features_out': 48,
            'expand_ratio': 4,
            'se_ratio': 0.0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'conv_type': 1,
            'expand_ratio': 4,
            'features_in': 48,
            'kernel_size': (3, 3),
            'num_repeat': 4,
            'features_out': 64,
            'se_ratio': 0,
            'strides': 2,
        },
        {
            'conv_type': 0,
            'expand_ratio': 4,
            'features_in': 64,
            'kernel_size': (3, 3),
            'num_repeat': 6,
            'features_out': 128,
            'se_ratio': 0.25,
            'strides': 2,
        },
        {
            'conv_type': 0,
            'expand_ratio': 6,
            'features_in': 128,
            'kernel_size': (3, 3),
            'num_repeat': 9,
            'features_out': 160,
            'se_ratio': 0.25,
            'strides': 1,
        },
        {
            'conv_type': 0,
            'expand_ratio': 6,
            'features_in': 160,
            'kernel_size': (3, 3),
            'num_repeat': 15,
            'features_out': 256,
            'se_ratio': 0.25,
            'strides': 2,
        },
    ],
    'efficientnetv2-m': [
        {
            'kernel_size': (3, 3),
            'num_repeat': 3,
            'features_in': 24,
            'features_out': 24,
            'expand_ratio': 1,
            'se_ratio': 0,
            'strides': 1,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 5,
            'features_in': 24,
            'features_out': 48,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 5,
            'features_in': 48,
            'features_out': 80,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 7,
            'features_in': 80,
            'features_out': 160,
            'expand_ratio': 4,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 14,
            'features_in': 160,
            'features_out': 176,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 1,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 18,
            'features_in': 176,
            'features_out': 304,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 5,
            'features_in': 304,
            'features_out': 512,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 1,
            'conv_type': 0,
        },
    ],
    'efficientnetv2-l': [
        {
            'kernel_size': (3, 3),
            'num_repeat': 4,
            'features_in': 32,
            'features_out': 32,
            'expand_ratio': 1,
            'se_ratio': 0,
            'strides': 1,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 7,
            'features_in': 32,
            'features_out': 64,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 7,
            'features_in': 64,
            'features_out': 96,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 10,
            'features_in': 96,
            'features_out': 192,
            'expand_ratio': 4,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 19,
            'features_in': 192,
            'features_out': 224,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 1,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 25,
            'features_in': 224,
            'features_out': 384,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 7,
            'features_in': 384,
            'features_out': 640,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 1,
            'conv_type': 0,
        },
    ],
    'efficientnetv2-b0': [
        {
            'kernel_size': (3, 3),
            'num_repeat': 1,
            'features_in': 32,
            'features_out': 16,
            'expand_ratio': 1,
            'se_ratio': 0,
            'strides': 1,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 2,
            'features_in': 16,
            'features_out': 32,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 2,
            'features_in': 32,
            'features_out': 48,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 3,
            'features_in': 48,
            'features_out': 96,
            'expand_ratio': 4,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 5,
            'features_in': 96,
            'features_out': 112,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 1,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 8,
            'features_in': 112,
            'features_out': 192,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
    ],
    'efficientnetv2-b1': [
        {
            'kernel_size': (3, 3),
            'num_repeat': 1,
            'features_in': 32,
            'features_out': 16,
            'expand_ratio': 1,
            'se_ratio': 0,
            'strides': 1,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 2,
            'features_in': 16,
            'features_out': 32,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 2,
            'features_in': 32,
            'features_out': 48,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 3,
            'features_in': 48,
            'features_out': 96,
            'expand_ratio': 4,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 5,
            'features_in': 96,
            'features_out': 112,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 1,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 8,
            'features_in': 112,
            'features_out': 192,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
    ],
    'efficientnetv2-b2': [
        {
            'kernel_size': (3, 3),
            'num_repeat': 1,
            'features_in': 32,
            'features_out': 16,
            'expand_ratio': 1,
            'se_ratio': 0,
            'strides': 1,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 2,
            'features_in': 16,
            'features_out': 32,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 2,
            'features_in': 32,
            'features_out': 48,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 3,
            'features_in': 48,
            'features_out': 96,
            'expand_ratio': 4,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 5,
            'features_in': 96,
            'features_out': 112,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 1,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 8,
            'features_in': 112,
            'features_out': 192,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
    ],
    'efficientnetv2-b3': [
        {
            'kernel_size': (3, 3),
            'num_repeat': 1,
            'features_in': 32,
            'features_out': 16,
            'expand_ratio': 1,
            'se_ratio': 0,
            'strides': 1,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 2,
            'features_in': 16,
            'features_out': 32,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 2,
            'features_in': 32,
            'features_out': 48,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 3,
            'features_in': 48,
            'features_out': 96,
            'expand_ratio': 4,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 5,
            'features_in': 96,
            'features_out': 112,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 1,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 8,
            'features_in': 112,
            'features_out': 192,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
    ],
}
